use std::fmt::{Debug, Display};

use binrw::{BinRead, BinWrite};
use diff::Diff;
use num::abs;

use crate::{
    field_of::FieldOf,
    types::{Coordinate, ObjectType, F64, U32},
};

use super::{ObjectCore, ObjectModifier, Translate};

#[derive(BinRead, BinWrite, Clone, Debug, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
pub struct Ellipse {
    pub core: ObjectCore,
    #[brw(magic(8u32))] // Number of following fields in struct
    pub clockwise: U32,
    pub drawn_corner_a: FieldOf<Coordinate>,
    pub drawn_corner_b: FieldOf<Coordinate>,
    pub start_angle: F64, // Radians
    pub end_angle: F64,   // Radians
    pub modifier: ObjectModifier,
    pub open_curve: U32,
}

impl Display for Ellipse {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}, Origin: {}, Width: {:.2}, Height: {:.2}, Start Radian: {:.2}, End Radian: {:.2}, Open Curve: {}", 
            self.core,
            Coordinate {
                x: self.modifier.modifiers.correction.x
                    + (self.drawn_corner_a.x + self.drawn_corner_b.x) / 2.0
                        * self.modifier.modifiers.x_scale,
                y: self.modifier.modifiers.correction.y
                    + (self.drawn_corner_a.y + self.drawn_corner_b.y) / 2.0
                        * self.modifier.modifiers.y_scale,
            },
            (abs(self.drawn_corner_a.x) + abs(self.drawn_corner_b.x)) / 2.0
                * self.modifier.modifiers.x_scale,
            (abs(self.drawn_corner_a.y) + abs(self.drawn_corner_b.y)) / 2.0
                * self.modifier.modifiers.y_scale,
                self.start_angle,
                self.end_angle,
                *self.open_curve != 0,
        )
    }
}

impl Default for Ellipse {
    fn default() -> Self {
        Self {
            core: ObjectCore::default(ObjectType::Ellipse),
            clockwise: 0.into(),
            drawn_corner_a: Coordinate { x: 0.0, y: 0.0 }.into(),
            drawn_corner_b: Coordinate { x: 0.0, y: 0.0 }.into(),
            start_angle: 0.0.into(),
            end_angle: 0.0.into(),
            modifier: ObjectModifier::default(),
            open_curve: 0.into(),
        }
    }
}

impl Translate for Ellipse {
    fn move_absolute(&mut self, origin: Option<Coordinate>, z: Option<f64>) {
        origin.map(|origin| {
            let delta: Coordinate = origin - *self.core.origin;
            self.modifier.move_relative(delta);
        });
        self.core.move_absolute(origin, z);
    }

    fn move_relative(&mut self, delta: Option<Coordinate>, z: Option<f64>) {
        delta.map(|delta| {
            self.modifier.move_relative(delta);
        });
        self.core.move_relative(delta, z);
    }
}
